from __future__ import annotations

import traceback
from functools import lru_cache, wraps
from inspect import isfunction
from os import makedirs
from os.path import abspath
from sys import stderr
from typing import Callable, Optional
from weakref import ref

from danling.typing import Exceptions


def flexible_decorator(maybe_decorator: Optional[Callable] = None):
    r"""
    Decorator to allow bracket-less when no arguments are passed.

    Examples:
        For decorator defined as follows:

        >>> @flexible_decorator
        ... def decorator(*args, **kwargs):
        ...     def wrapper(func, *args, **kwargs):
        ...         pass
        ...     return wrapper

        The following two are equivalent:

        >>> @decorator
        ... def func(*args, **kwargs):
        ...     pass

        >>> @decorator()
        ... def func(*args, **kwargs):
        ...     pass
    """

    def decorator(func: Callable):
        @wraps(decorator)
        def wrapper(*args, **kwargs):
            if len(args) == 1 and isfunction(args[0]):
                return func(**kwargs)(args[0])
            return func(*args, **kwargs)

        return wrapper

    if maybe_decorator is None:
        return decorator
    return decorator(maybe_decorator)


def print_exc(exc, func, args, kwargs, verbosity: int = 40):  # pylint: disable=W0613
    r"""
    Print exception raised by `func` with `args` and `kwargs` to `stderr`.
    This function serves as the default callback for catch.

    Args:
        verbosity: What level of traceback to print.
            0-: No traceback.
            0-10: Full information of arguments and key word arguments.
            10-20: Stack trace to function calls.
            40+: Function name and error messages.
    """

    if verbosity >= 0:
        message = traceback.format_exc()
        message += f"\nencoutered when calling {func}"
        if verbosity <= 20:
            message += "\n\nstack:\n" + "\n".join(traceback.format_stack()[:-2])
        if verbosity <= 10:
            message += "\n" + f"args: {args}\nkwargs: {kwargs}"
        try:
            print(message, file=stderr, force=True)  # type: ignore
        except TypeError:
            print(message, file=stderr)


@flexible_decorator
def catch(  # pylint: disable=W1113
    error: Exceptions = Exception,
    exclude: Exceptions | None = None,
    callback: Callable = print_exc,
    *callback_args,
    **callback_kwargs,
):
    r"""
    Decorator to catch `error` except for `exclude`.
    Detailed traceback will be printed to `stderr`.

    `catch` is extremely useful for unfatal errors.
    For example, `Runner` saves checkpoint regularly, however, this might break running if the space is full.
    Decorating `save` method with `catch` will allow you to catch these errors and continue your running.

    Args:
        error: Exceptions to be caught.
        exclude: Exceptions to be excluded.
        callback: Callback to be called when an error occurs.
            The first four arguments to `callback` are `exc`, `func`, `args`, `kwargs`.
            Additional arguments should be passed with `*callback_args` and `**callback_kwargs`.
        callback_args: Arguments to be passed to `callback`.
        callback_kwargs: Keyword arguments to be passed to `callback`.

    Examples:
        >>> def file_not_found(*args, **kwargs):
        ...     return exclude
        ...     raise FileNotFoundError

        >>> file_not_found()
        >>> raise ValueError
        >>> assert 1 == 2
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):  # pylint: disable=R1710
            try:
                return func(*args, **kwargs)
            except error as exc:  # pylint: disable=W0703
                if exclude is not None and isinstance(exc, exclude):
                    raise exc
                callback(exc, func, args, kwargs, *callback_args, **callback_kwargs)

        return wrapper

    decorator.__doc__ = catch.__doc__

    return decorator


def method_cache(*cache_args, **lru_kwargs):
    r"""
    Decorator to cache the result of an instance method.

    `functools.lru_cache` uses a strong reference to the instance,
    which will make the instance immortal and break the garbage collection.

    `method_cache` uses a weak reference to the instance and works fine.

    https://rednafi.github.io/reflections/dont-wrap-instance-methods-with-functoolslru_cache-decorator-in-python.html
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            self_ref = ref(self)

            @wraps(func)
            @lru_cache(*cache_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_ref(), *args, **kwargs)

            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)

        return wrapper

    return decorator


def ensure_dir(func):
    r"""
    Decorator to ensure a directory property exists.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        path = abspath(func(*args, **kwargs))
        makedirs(path, exist_ok=True)
        return path

    return wrapper



# pylint: disable=E1101,W0622

from __future__ import annotations

import torch
from torch import Tensor
from torcheval.metrics import functional as tef
from torchmetrics import functional as tmf

from danling.tensors import NestedTensor


def auroc(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
    weight: Tensor | None = None,
    average: str | None = "macro",
    num_labels: int | None = None,
    num_classes: int | None = None,
    task_weights: Tensor | None = None,
    **kwargs
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    if num_labels is None and num_classes is None:
        return tef.binary_auroc(input=input, target=target, weight=weight, **kwargs)
    if num_classes is None:
        ret = tef.binary_auroc(input=input.T, target=target.T, num_tasks=num_labels, weight=weight, **kwargs)
        if task_weights is not None:
            return ret @ task_weights.double()
        return ret.mean()
    if num_labels is None:
        return tef.multiclass_auroc(input=input, target=target, num_classes=num_classes, average=average, **kwargs)
    raise ValueError("Could not infer the type of the task. Only one of `num_labels`, `num_classes` is allowed.")


def auprc(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
    average: str | None = "macro",
    num_labels: int | None = None,
    num_classes: int | None = None,
    task_weights: Tensor | None = None,
    **kwargs
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    if num_labels is None and num_classes is None:
        return tef.binary_auprc(input=input, target=target, **kwargs)
    if num_classes is None:
        ret = tef.multilabel_auprc(input=input, target=target, num_labels=num_labels, average=average, **kwargs)
        if task_weights is not None:
            return ret @ task_weights.double()
        return ret.mean()
    if num_labels is None:
        return tef.multiclass_auprc(input=input, target=target, num_classes=num_classes, average=average, **kwargs)
    raise ValueError("Could not infer the type of the task. Only one of `num_labels`, `num_classes` is allowed.")


def accuracy(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
    threshold: float = 0.5,
    average: str | None = "micro",
    num_labels: int | None = None,
    num_classes: int | None = None,
    **kwargs
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    if num_labels is None and num_classes is None:
        return tef.binary_accuracy(input=input, target=target, threshold=threshold, **kwargs)
    if num_classes is None:
        return tef.multilabel_accuracy(input=input, target=target, threshold=threshold, **kwargs)
    if num_labels is None:
        return tef.multiclass_accuracy(input=input, target=target, num_classes=num_classes, average=average, **kwargs)
    raise ValueError("Could not infer the type of the task. Only one of `num_labels`, `num_classes` is allowed.")


def pearson(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    return tmf.pearson_corrcoef(input, target)


def spearman(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    return tmf.spearman_corrcoef(input, target)


def r2_score(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    return tef.r2_score(input, target)


def mse(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    return tef.mean_squared_error(input, target)


def rmse(
    input: Tensor | NestedTensor,
    target: Tensor | NestedTensor,
):
    if isinstance(input, NestedTensor):
        input = torch.cat(input.storage)
    if isinstance(target, NestedTensor):
        target = torch.cat(target.storage)
    return tef.mean_squared_error(input, target).sqrt()



import torchvision
from chanfig import Config, Registry
from torch import nn, optim

import danling as dl

OPTIMIZERS = Registry()
OPTIMIZERS.register(optim.AdamW, "adamw")
OPTIMIZERS.register(optim.SGD, "sgd")


class MNISTConfig(Config):
    def __init__(self):
        self.network.name = "resnet18"
        self.dataset.download = True
        self.dataset.root = "data"
        self.dataloader.batch_size = 8
        self.epoch_end = 2
        self.optim.name = "adamw"
        self.optim.lr = 1e-3
        self.optim.weight_decay = 1e-4
        self.log = False
        self.tensorboard = False
        self.print_interval = 1000
        self.score_set = "val"
        self.score_name = "loss"
        self.debug = False
        self.patience = 1

    def post(self):
        self.experiment_name = f"{self.network.name}_{self.optim.name}@{self.optim.lr}"


class MNISTRunner(dl.TorchRunner):
    def __init__(self, config: Config):
        super().__init__(config)

        self.dataset.transform = torchvision.transforms.Compose(
            [
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )
        self.datasets.train = torchvision.datasets.MNIST(train=True, **self.dataset)
        self.datasets.val = torchvision.datasets.MNIST(train=False, **self.dataset)

        self.model = getattr(torchvision.models, self.network.name)(pretrained=False, num_classes=10)
        self.model.conv1 = nn.Conv2d(1, 64, 1, bias=False)
        self.optimizer = OPTIMIZERS.build(params=self.model.parameters(), **self.optim)
        self.criterion = nn.CrossEntropyLoss()

        self.meters.loss.reset()
        self.meters.time.reset()


if __name__ == "__main__":
    config = MNISTConfig()
    config.parse()
    with dl.debug(config.get("debug", False)):
        runner = MNISTRunner(config)
        runner.train()
        runner.evaluate()



from __future__ import annotations

import os
import random
from collections.abc import Callable, Mapping
from contextlib import suppress
from time import time
from typing import Any
from warnings import warn

import torch
from accelerate import Accelerator
from accelerate.utils import DeepSpeedPlugin
from chanfig import NestedDict
from torch import distributed as dist
from torch import nn, optim, utils
from torch.backends import cudnn
from tqdm import tqdm

try:
    from numpy import random as np_random
except ImportError:
    np_random = None

from danling.utils import catch

from .base_runner import BaseRunner
from .utils import on_main_process


class TorchRunner(BaseRunner):
    r"""
    Set up everything for running a job.

    `TorchRunner` uses [`accelerate`][accelerate] as distributed backend to
    provide seamless distributed training experience.

    `TorchRunner` will automatically [`prepare`][accelerate.Accelerator.prepare] everything,
    including `model`, `criterion`, `optimizer`, `scheduler`, and `dataloaders` for distribute training,
    mixed precision, and deepspeed (optional).

    In fact, you don't even need to create `dataloaders`, just define
    `datasets` and `TorchRunner` will create `dataloaders` for you.
    `TorchRunner` will inspect the `train` flag in corresponding dataset to
    automatically set `shuffle`.

    Attributes:
        accelerator (Accelerator):
        accelerate: Arguments to pass when building accelerator. Defaults to `{}`.
    """

    # pylint: disable=R0902

    accelerator: Accelerator
    accelerate: dict

    model: nn.Module
    criterion: nn.Module
    optimizer: optim.Optimizer
    scheduler: optim.lr_scheduler._LRScheduler

    def __init__(self, *args, **kwargs) -> None:
        if len(args) != 1 or kwargs:
            message = (
                "Passing multiple args & kwargs to build Runner is deprecated and will be removed in DanLing v0.3.\n"
                "Please only pass a config dict instead."
            )
            warn(message, DeprecationWarning, stacklevel=2)
            config = NestedDict(*args, **kwargs)
        else:
            config = args[0]
        if "accelerate" not in self:  # class attributes
            self.accelerate = {}
        self.accelerate.update(config.get("accelerate", {}))
        super().__init__(config)

    def __post_init__(self, *args, **kwargs) -> None:
        self._prepare()

    def _prepare(self):
        if self.datasets:
            datasets = {k: d for k, d in self.datasets.items() if k not in self.dataloaders}
            dataloader_kwargs = self.state.get("dataloader", {})
            for k, d in datasets.items():
                shuffle = dataloader_kwargs.shuffle if "shuffle" in dataloader_kwargs else getattr(d, "train", True)
                self.dataloaders[k] = utils.data.DataLoader(d, shuffle=shuffle, **dataloader_kwargs)
        objects = [self.model, self.criterion, self.optimizer, self.scheduler]
        num_objects = len(objects)
        dataloader_names = []
        for name, dataloader in self.dataloaders.items():
            dataloader_names.append(name)
            objects.append(dataloader)
        objects = self.prepare(*objects)
        self.model, self.criterion, self.optimizer, self.scheduler = objects[:num_objects]
        if len(objects) != len(dataloader_names) + num_objects:
            raise ValueError("Number of dataloaders does not match.")
        for name, dataloader in zip(dataloader_names, objects[num_objects:]):
            self.dataloaders[name] = dataloader

    @property
    def deepspeed(self) -> dict | None:
        if "accelerator" not in self:
            raise ValueError("accelerator is not used")
        if self.accelerator.state.deepspeed_plugin is not None:
            return self.accelerator.state.deepspeed_plugin.deepspeed_config
        return None

    def train(self, train_splits: list[str] | None = None, eval_splits: list[str] | None = None) -> NestedDict:
        r"""
        Perform training on `split`.

        Args:
            train_splits (list[str]): list of split to run train.
                Defaults to `["train"]`.
            eval_splits (list[str]): list of split to run evaluate.
                Defaults to `self.dataloaders` except for those in `train_splits`.

        Return:
            NestedDict: train results
        """

        early_stop_counter = 0
        if train_splits is None:
            train_splits = ["train"]
        if eval_splits is None:
            eval_splits = [s for s in self.dataloaders if s not in train_splits]
        self.state.epoch_begin = self.state.epochs
        print(f"Begin training from {self.state.epoch_begin} to {self.state.epoch_end}")
        print(f"Training splits: {train_splits}")
        print(f"Evaluation splits: {eval_splits}")
        patience = self.state.get("patience", float("inf"))
        for epochs in range(self.state.epoch_begin, self.state.epoch_end):  # type: ignore
            self.state.epochs = epochs
            result = NestedDict()
            result.setattr("convert_mapping", True)
            for split in train_splits:
                result[split] = self.train_epoch(split)
            for split in eval_splits:
                result[split] = self.evaluate_epoch(split)
            self.append_result(result)
            print(self.format_epoch_result(result))
            self.save_result()
            self.save_checkpoint()
            """@nni.report_intermediate_result(self.latest_score)"""  # pylint: disable=W0105
            early_stop_counter = 0 if self.is_best else early_stop_counter + 1
            if early_stop_counter > patience:
                print("early stop")
                break
        """@nni.report_final_result(self.latest_score)"""  # pylint: disable=W0105
        return self.results

    def train_epoch(self, split: str = "train") -> NestedDict:
        r"""
        Train one epoch on `split`.

        Args:
            split (str): split to run train

        Return:
            NestedDict: train result
        """

        # pylint: disable=E1101, E1102, W0622
        self.mode = "train"  # type: ignore
        loader = self.dataloaders[split]
        length = len(loader) - 1
        self.meters.reset()
        if self.metrics is not None:
            self.metrics.reset()
        batch_time = time()
        if hasattr(loader.batch_sampler, "set_epoch"):
            loader.batch_sampler.set_epoch(self.epochs)
        if hasattr(loader.sampler, "set_epoch"):
            loader.sampler.set_epoch(self.epochs)

        for iteration, data in enumerate(loader):
            with self.autocast(), self.accumulate():
                input = data["input"] if isinstance(data, Mapping) else data[0]
                target = data["target"] if isinstance(data, Mapping) else data[1]
                pred = self.model(**input) if isinstance(input, Mapping) else self.model(input)
                loss = self.criterion(pred, target)
                if self.metrics is not None:
                    self.metrics.update(pred, target)
                self.step(loss)

            if self.print_interval > 0 and iteration % self.print_interval == 0:
                if self.device == torch.device("cuda"):
                    torch.cuda.synchronize()
                self.meters.time.update((time() - batch_time) / self.print_interval)
                batch_time = time()
                reduced_loss = self.reduce(loss).item()
                self.meters.loss.update(reduced_loss)
                self.step_log(split, iteration, length)

        result = self.meters.avg
        if self.metrics is not None:
            result.merge(self.metrics.avg)
        return result

    def evaluate(self, eval_splits: list[str] | None = None) -> NestedDict:
        r"""
        Perform evaluation on `eval_splits`.

        Args:
            eval_splits (list[str]): list of split to run evaluate.
                Defaults to `["eval"]`.

        Return:
            NestedDict: evaluation result
        """

        if eval_splits is None:
            eval_splits = ["eval"]

        print("Begin evaluation")
        print(f"Evaluation splits: {eval_splits}")
        result = NestedDict()
        result.setattr("convert_mapping", True)
        for split in eval_splits:
            result[split] = self.evaluate_epoch(split=split)
        print(self.format_epoch_result(result))
        return result

    @torch.inference_mode()
    def evaluate_epoch(self, split: str = "val") -> NestedDict:
        r"""
        Evaluate one epoch on `split`.

        Args:
            split (str): split to run evaluate

        Return:
            NestedDict: evaluation result
        """

        # pylint: disable=E1101, E1102, W0622
        self.mode = "eval"  # type: ignore
        loader = self.dataloaders[split]
        length = len(loader) - 1
        self.meters.reset()
        if self.metrics is not None:
            self.metrics.reset()
        batch_time = time()

        for iteration, data in enumerate(loader):
            input = data["input"] if isinstance(data, Mapping) else data[0]
            target = data["target"] if isinstance(data, Mapping) else data[1]
            pred = self.model(**input) if isinstance(input, Mapping) else self.model(input)
            loss = self.criterion(pred, target)
            if self.metrics is not None:
                self.metrics.update(pred, target)

            if self.print_interval > 0 and iteration % self.print_interval == 0:
                if self.device == torch.device("cuda"):
                    torch.cuda.synchronize()
                self.meters.time.update((time() - batch_time) / self.print_interval)
                batch_time = time()
                reduced_loss = self.reduce(loss).item()
                self.meters.loss.update(reduced_loss)
                self.step_log(split, iteration, length)

        result = self.meters.avg
        if self.metrics is not None:
            result.merge(self.metrics.avg)
        self.write_result(result, split, self.state.epochs)
        return result

    @torch.inference_mode()
    def inference(self, split: str = "inf") -> list:
        r"""
        Perform inference on `split`.

        Args:
            split (str): split to run inference

        Return:
            Tensor: inference outputs
        """

        # pylint: disable=E1102, W0622
        self.mode = "inf"  # type: ignore
        loader = self.dataloaders[split]
        self.meters.reset()
        output = []
        for _, data in tqdm(enumerate(loader), total=len(loader)):
            input = data["input"] if isinstance(data, Mapping) else data[0]
            pred = self.model(**input) if isinstance(input, Mapping) else self.model(input)
            output.extend(pred.tolist())

        if self.distributed:
            torch.cuda.synchronize()
            output = self.gather_for_metrics(output)
        return output

    def init_distributed(self) -> None:
        r"""
        Set up distributed training.

        Initialise process group and set up DDP variables.
        """

        if os.environ.get("ACCELERATE_USE_DEEPSPEED", "false").lower() == "true":
            deepspeed_config = self.state.get("deepspeed", os.environ.get("ACCELERATE_DEEPSPEED_CONFIG_FILE"))
            self.accelerate["deepspeed_plugin"] = DeepSpeedPlugin(hf_ds_config=self.init_deepspeed(deepspeed_config))
        self.accelerator = Accelerator(**self.accelerate)
        if self.distributed:
            object_list = [self.state.id]
            dist.broadcast_object_list(object_list)
            self.state.id = object_list[0]

    @on_main_process
    def init_tensorboard(self, *args, **kwargs) -> None:
        r"""
        Set up Tensoraoard SummaryWriter.
        """
        from torch.utils.tensorboard.writer import SummaryWriter  # pylint: disable=C0415

        if "log_dir" not in kwargs:
            kwargs["log_dir"] = self.dir

        self.writer = SummaryWriter(*args, **kwargs)
        self.writer.add_scalar = catch(OSError, verbose=False)(self.writer.add_scalar)

    def set_seed(self, seed: int | None = None, bias: int | None = None) -> None:
        r"""
        Set up random seed.

        Args:
            seed: Random seed to set.
                Defaults to `self.state.seed` (`config.seed`).

            bias: Make the seed different for each processes.
                This is used to ensure the data augmentation are applied differently on every processes.
                Defaults to `self.rank`.
                Set to `False` to disable this feature.
        """

        seed = seed or self.state.seed
        if self.distributed:
            object_list = [seed]
            dist.broadcast_object_list(object_list)
            seed = object_list[0]
        bias = bias or self.rank
        if bias:
            seed += bias
        self.state.seed = seed
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        if np_random is not None:
            np_random.seed(seed)
        random.seed(seed)

    def set_deterministic(self) -> None:
        r"""
        Set up deterministic.
        """

        cudnn.benchmark = False
        cudnn.deterministic = True
        if torch.__version__ >= "1.8.0":
            torch.use_deterministic_algorithms(True)

    def step(self, loss, batch_size: int | None = None, zero_grad: bool = True) -> None:
        r"""
        Backward loss and step optimizer & scheduler.

        This method increment `self.state.steps`.

        This method also increment `self.state.iters` when `batch_size` is specified.

        Args:
            zero_grad: Whether to zero the gradients.
        """

        self.accelerator.backward(loss)
        if self.sync_gradients:
            if self.state.get("max_grad_value") is not None:
                self.clip_grad_value_(self.model.parameters(), self.state.get("max_grad_value"))  # type: ignore
            if self.state.get("max_grad_norm") is not None:
                self.clip_grad_norm_(self.model.parameters(), self.state.get("max_grad_norm"))  # type: ignore
        if self.optimizer is not None:
            self.optimizer.step()
            if zero_grad:
                self.optimizer.zero_grad()
        if self.scheduler is not None:
            self.scheduler.step()
        self.state.steps += 1
        if batch_size is None:
            batch_size = self.batch_size_equivalent
        self.state.iters += batch_size
        # TODO: Support `drop_last = False`
        # self.state.iters += self.batch_size_equivalent

    def state_dict(self, cls: Callable = dict) -> Mapping:
        r"""
        Return dict of all attributes for checkpoint.
        """

        if self.model is None:
            raise ValueError("Model must be defined when calling state_dict")
        model = self.accelerator.unwrap_model(self.model)
        return cls(
            runner=self.state.dict(),
            model=model.state_dict(),
            optimizer=self.optimizer.state_dict() if self.optimizer else None,
            scheduler=self.scheduler.state_dict() if self.scheduler else None,
        )

    def prepare(self, *args, device_placement: list[bool] | None = None) -> None:
        r"""
        Prepare all objects passed in `args` for distributed training and mixed precision,
        then return them in the same order.
        """

        return self.accelerator.prepare(*args, device_placement=device_placement)

    def accumulate(self, model: nn.Module | None = None):
        r"""
        Context manager that enables gradient accumulate.
        """

        model = model or self.model
        return self.accelerator.accumulate(model)

    def autocast(self):
        r"""
        Context manager that enables auto-casting for the forward pass (and maybe backward pass).
        """

        return self.accelerator.autocast()

    def backward(self, loss) -> None:
        r"""
        Backward loss to compute gradients.
        """

        return self.accelerator.backward(loss)

    def unwrap_model(self, model: nn.Module | None = None) -> nn.Module:
        r"""
        Unwrap DDP model.

        Args:
            model (Optional[nn.Module]):
                Defaults to `self.model`.
        """

        if model is not None:
            model = self.model
        if self.accelerator is not None:
            return self.accelerator.unwrap_model(model)
        if self.distributed:
            return model.module
        return model

    @property
    def batch_size(self) -> int:
        r"""
        Batch size.

        Notes:
            If `train` is in `dataloaders`, then `batch_size` is the batch size of `train`.
            Otherwise, `batch_size` is the batch size of the first dataloader.

        Returns:
            (int):
        """

        batch_size = self.state.get("dataloader.batch_size")
        if batch_size:
            return batch_size
        if self.dataloaders:
            loader = self.dataloaders.get("train", next(iter(self.dataloaders.values())))
            if loader.batch_size:
                return loader.batch_size
            batch_sampler = loader.batch_sampler if loader.batch_sampler is not None else loader.sampler
            return batch_sampler.batch_size
        raise AttributeError("batch_size could not be inferred, since no dataloader found.")

    @property
    def accum_steps(self) -> int:
        r"""
        Gradient accumulation steps.

        Returns:
            (int):
        """

        return self.accelerator.gradient_accumulation_steps

    @property
    def device(self) -> torch.device:  # pylint: disable=E1101
        r"""
        Device of runner.
        """

        return self.accelerator.device

    @property
    def world_size(self) -> int:
        r"""
        Number of Processes.
        """

        return self.accelerator.num_processes

    @property
    def rank(self) -> int:
        r"""
        Process index in all processes.
        """

        return self.accelerator.process_index

    @property
    def local_rank(self) -> int:
        r"""
        Process index in local processes.
        """

        return self.accelerator.local_process_index

    def gather(self, tensor) -> torch.Tensor:
        r"""
        Gather tensor.
        """

        return self.accelerator.gather(tensor)

    def reduce(self, tensor, reduction: str = "sum") -> torch.Tensor:
        r"""
        Reduce tensor.
        """

        return self.accelerator.reduce(tensor, reduction=reduction)

    def __getattr__(self, name: str) -> Any:
        with suppress(AttributeError):
            return super().__getattr__(name)
        if "accelerator" in self.__dict__ and hasattr(self.accelerator, name):
            return getattr(self.accelerator, name)
        raise super().__getattribute__(name)



import pytest

from danling.utils import decorators


class Test:
    def test_catch_error(self):
        @decorators.catch()
        def func():
            raise Exception("test")

        func()

    def test_catch_interrupt(self):
        @decorators.catch()
        def func():
            raise KeyboardInterrupt("test")

        with pytest.raises(KeyboardInterrupt):
            func()

    def test_catch_exit(self):
        @decorators.catch()
        def func():
            raise SystemExit("test")

        with pytest.raises(SystemExit):
            func()

    def test_catch_raise(self):
        @decorators.catch(FileExistsError)
        def func():
            raise FileNotFoundError("test")

        with pytest.raises(FileNotFoundError):
            func()

    def test_catch_exclude(self):
        @decorators.catch(exclude=FileNotFoundError)
        def func():
            raise FileNotFoundError("test")

        with pytest.raises(FileNotFoundError):
            func()



from functools import partial

from .average_meters import AverageMeter, AverageMeters
from .functional import accuracy, auprc, auroc, pearson, r2_score, rmse, spearman
from .metrics import Metrics

__all__ = [
    "Metrics",
    "AverageMeter",
    "AverageMeters",
    "regression_metrics",
    "binary_metrics",
    "multiclass_metrics",
    "multilabel_metrics",
]


def binary_metrics():
    return Metrics(auroc=auroc, auprc=auprc, acc=accuracy)


def multiclass_metrics(num_classes: int):
    p_auroc = partial(auroc, num_classes=num_classes)
    p_auprc = partial(auprc, num_classes=num_classes)
    p_acc = partial(accuracy, num_classes=num_classes)
    return Metrics(auroc=p_auroc, auprc=p_auprc, acc=p_acc)


def multilabel_metrics(num_labels: int):
    p_auroc = partial(auroc, num_labels=num_labels)
    p_auprc = partial(auprc, num_labels=num_labels)
    return Metrics(auroc=p_auroc, auprc=p_auprc, acc=accuracy)


def regression_metrics():
    return Metrics(
        pearson=pearson,
        spearman=spearman,
        r2=r2_score,
        rmse=rmse,
    )



DEFAULT_RUN_NAME = "Run"
DEFAULT_EXPERIMENT_NAME = "DanLing"
DEFAULT_EXPERIMENT_ID = "xxxxxxxxxxxxxxxx"
DEFAULT_IGNORED_KEYS_IN_HASH = {
    "iters",
    "steps",
    "epochs",
    "results",
    "index_set",
    "index",
}



from __future__ import annotations

import logging
import logging.config
import os
import random
import shutil
from collections.abc import Callable, Mapping, Sequence
from math import ceil
from sys import version_info
from typing import Any
from warnings import warn

from chanfig import Config, FlatDict, NestedDict, Variable

from danling.metrics import AverageMeter, AverageMeters, Metrics
from danling.typing import File, PathStr
from danling.utils import catch, ensure_dir, load, save

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property  # type: ignore

try:
    from numpy import random as np_random
except ImportError:
    np_random = None

from .runner_state import RunnerState
from .utils import RunnerMeta, RunnerMode, on_main_process

PY38_PLUS = version_info >= (3, 8)
IGNORED_SET_NAMES = ("index", "epoch", "step", "iter")
__APPEND_RESULT_COUNTER__ = 0


class BaseRunner(metaclass=RunnerMeta):
    r"""
    Base class for all runners.

    `BaseRunner` sets up basic running environment, including `seed`, `deterministic`, and `logging`.

    `BaseRunner` also provides some basic methods, such as, `step`, `state_dict`, `save_checkpoint`, `load_checkpoint`.

    `BaseRunner` defines all basic attributes and relevant properties such as `scores`, `progress`, etc.

    Attributes: Core:
        mode (RunnerMode, property): Running mode.
        state (RunnerState): Running state. See `RunnerState` for details.

    Attributes: Model:
        model (Callable):
        criterion (Callable):
        optimizer:
        scheduler:

    Attributes: Data:
        datasets (FlatDict): All datasets, should be in the form of ``{subset: dataset}``.
            Initialised to `FlatDict` by default.
        datasamplers (FlatDict): All datasamplers, should be in the form of ``{subset: datasampler}``.
            Initialised to `FlatDict` by default.
        dataloaders (FlatDict): All dataloaders, should be in the form of ``{subset: dataloader}``.
            Initialised to `FlatDict` by default.
        batch_size (int, property): Number of samples per batch in train dataloader or the first dataloader.
        batch_size_equivalent (int, property): Total batch_size (`batch_size * world_size * accum_steps`).

    `datasets`, `datasamplers`, `dataloaders` should be a dict with the same keys.
    Their keys should be `split` (e.g. `train`, `val`, `test`).

    Attributes: Progress:
        progress (float, property): Running Progress, in `range(0, 1)`.

    Attributes: Results:
        results (NestedDict): Results include all metric information of the model.
            Results should be in the form of `{epoch: {subset: {metric: score}}}`.
        latest_result (NestedDict, property): Most recent result, should be in the form of `{subset: {metric: score}}`.
        best_result (NestedDict, property): Best result, should be in the form of `{subset: {metric: score}}`.
        scores (List[float], property): Score is the core metric that is used to evaluate the performance of the model.
            Scores should be in the form of `{epoch: score}`.
        latest_score (float, property): Most recent score, should be in the form of `score`.
        best_score (float, property): Best score, should be in the form of `score`.
        score_set (Optional[str]): The subset to calculate the score.
            If is `None`, will use the last set of the result.
        score_name (str): The metric name of score.
            Defaults to `"loss"`.
        is_best (bool, property): If `latest_score == best_score`.

    Attributes: IO:
        dir (str, property): Directory of the run.
            Defaults to `os.path.join(self.project_root, f"{self.name}-{self.id}")`.
        checkpoint_dir (str, property): Directory of checkpoints.
        log_path (str, property):  Path of log file.
        checkpoint_dir_name (str): The name of the directory under `runner.dir` to save checkpoints.
            Defaults to `"checkpoints"`.

    Attributes: Parallel Training:
        world_size (int, property): Number of processes.
        rank (int, property): Process index of all processes.
        local_rank (int, property): Process index of local processes.
        distributed (bool, property): If runner is running in distributed mode.
        is_main_process (bool, property): If current process is the main process of all processes.
        is_local_main_process (bool, property): If current process is the main process of local processes.

    Attributes: logging:
        meters (AverageMeters): Average meters.
            Initialised to `AverageMeters` by default.
        metrics (Metrics): Metrics for evaluating.
        logger:
        writer:

    See Also:
        [`RunnerState`][danling.runner.runner_state.RunnerState]: The runeer base that stores runtime information.
        [`BaseRunner`][danling.runner.BaseRunner]: The base runner class.
    """

    # pylint: disable=R0902, R0904
    # DO NOT set default value in class, as they won't be stored in `__dict__`.

    _mode: RunnerMode
    state: RunnerState

    model: Callable | None = None
    criterion: Callable | None = None
    optimizer: Any | None = None
    scheduler: Any | None = None

    datasets: FlatDict
    datasamplers: FlatDict
    dataloaders: FlatDict

    meters: AverageMeters
    metrics: Metrics | None = None
    logger: logging.Logger | None = None
    writer: Any | None = None

    def __init__(self, config: NestedDict) -> None:
        if "datasets" not in self.__dict__:
            self.datasets = FlatDict()
        if "datasamplers" not in self.__dict__:
            self.datasamplers = FlatDict()
        if "dataloaders" not in self.__dict__:
            self.dataloaders = FlatDict()
        self._mode = RunnerMode.train
        self.meters = AverageMeters()
        self.metrics = None
        # must init state at last to avoid conflict names
        self.state = RunnerState(config)
        self.init_distributed()
        if self.state.seed is not None:
            self.set_seed()
        if self.state.deterministic:
            self.set_deterministic()
        if os.listdir(self.dir):
            warn(
                f"Directory `{self.dir}` is not empty.",
                category=RuntimeWarning,
                stacklevel=2,
            )
        if self.state.log:
            self.init_logging()
        self.init_print()
        if self.state.tensorboard:
            self.init_tensorboard()

    def __post_init__(self, *args, **kwargs) -> None:
        pass

    @property
    def mode(self) -> RunnerMode:
        return self._mode

    @mode.setter
    def mode(self, mode: str | RunnerMode) -> None:
        if isinstance(mode, str):
            mode = RunnerMode(mode)
        self._mode = mode
        if self.model is not None:
            self.model.train(mode == RunnerMode.train)  # type: ignore

    @cached_property
    def batch_size(self) -> int:
        r"""
        Batch size.

        Notes:
            If `train` is in `dataloaders`, then `batch_size` is the batch size of `train`.
            Otherwise, `batch_size` is the batch size of the first dataloader.

        Returns:
            (int):
        """

        batch_size = self.state.get("dataloader.batch_size")
        if batch_size:
            return batch_size
        if self.dataloaders:
            loader = self.dataloaders["train"] if "train" in self.dataloaders else next(iter(self.dataloaders.values()))
            return loader.batch_size
        raise AttributeError("batch_size could not be inferred, since no dataloader found.")

    @property
    def batch_size_equivalent(self) -> int:
        r"""
        Actual batch size.

        Returns:
            (int): `batch_size` * `world_size` * `accum_steps`
        """

        return self.batch_size * self.world_size * getattr(self, "accum_steps", 1)

    @cached_property
    def total_epochs(self) -> int:
        if self.state.epoch_end:
            return self.state.epoch_end - self.state.epoch_begin
        raise ValueError("epoch_end is not specified")

    @cached_property
    def total_steps(self) -> int:
        if self.state.step_end:
            return self.state.step_end - self.state.step_begin
        dataset = self.datasets.get("train", next(iter(self.datasets.values())))
        return self.total_epochs * ceil(len(dataset) / self.batch_size)

    @cached_property
    def accum_steps(self) -> int:
        r"""
        Accumulated steps.

        Returns:
            (int):
        """

        return self.state.get("accum_steps", 1)

    def init_distributed(self) -> None:
        r"""
        Initialise distributed running environment.
        """

        raise NotImplementedError

    @property
    def device(self) -> Any:
        r"""
        Device of runner.
        """

        return "cpu"

    @property
    def world_size(self) -> int:
        r"""
        Number of processes.
        """

        return 1

    @property
    def rank(self) -> int:
        r"""
        Process index of all processes.
        """

        return 0

    @property
    def local_rank(self) -> int:
        r"""
        Process index of local processes.
        """

        return 0

    @property
    def distributed(self) -> bool:
        r"""
        If runner is running in distributed mode.
        """

        return self.world_size > 1

    @property
    def is_main_process(self) -> bool:
        r"""
        If current process is the main process of all processes.
        """

        return self.rank == 0

    @property
    def is_local_main_process(self) -> bool:
        r"""
        If current process is the main process of local processes.
        """

        return self.local_rank == 0

    @catch
    def save(  # pylint: disable=W1113
        self, obj: Any, file: PathStr, main_process_only: bool = True, *args, **kwargs
    ) -> File:
        r"""
        Save any file with supported extensions.

        `Runner.save` internally calls `dl.save`,
        but with additional arguments to allow it save only on the main process.
        Moreover, any error raised by `Runner.save` will be caught and logged.
        """

        if main_process_only and self.is_main_process or not main_process_only:
            return save(obj, file, *args, **kwargs)
        return file

    @staticmethod
    def load(file: PathStr, *args, **kwargs) -> Any:  # pylint: disable=C0103
        r"""
        Load any file with supported extensions.

        `Runner.load` is identical to `dl.load`.
        """

        return load(file, *args, **kwargs)

    def dict(self, cls: Callable = dict) -> Mapping:
        r"""
        Convert state to Mapping.

        Args:
            cls: Target `clc to convert to.
        """

        # pylint: disable=C0103

        return self.state.dict(cls)

    @catch
    def json(self, file: File, main_process_only: bool = True, *args, **kwargs) -> None:  # pylint: disable=W1113
        r"""
        Dump Runner State to json file.
        """

        if main_process_only and self.is_main_process or not main_process_only:
            return self.state.json(file, *args, **kwargs)

    @classmethod
    def from_json(cls, file: File, *args, **kwargs) -> BaseRunner:
        r"""
        Construct Runner from json file.

        This function calls `self.from_jsons()` to construct object from json string.
        You may overwrite `from_jsons` in case something is not json serializable.
        """

        with FlatDict.open(file) as fp:  # pylint: disable=C0103
            return cls.from_jsons(fp.read(), *args, **kwargs)

    def jsons(self, *args, **kwargs) -> str:
        r"""
        Dump Runner State to json string.
        """

        return self.state.jsons(*args, **kwargs)

    @classmethod
    def from_jsons(cls, string: str, *args, **kwargs) -> BaseRunner:
        r"""
        Construct Runner from json string.
        """

        return cls(Config.from_jsons(string, *args, **kwargs))

    @catch
    def yaml(self, file: File, main_process_only: bool = True, *args, **kwargs) -> None:  # pylint: disable=W1113
        r"""
        Dump Runner State to yaml file.
        """

        if main_process_only and self.is_main_process or not main_process_only:
            return self.state.yaml(file, *args, **kwargs)

    @classmethod
    def from_yaml(cls, file: File, *args, **kwargs) -> BaseRunner:
        r"""
        Construct Runner from yaml file.

        This function calls `self.from_yamls()` to construct object from yaml string.
        You may overwrite `from_yamls` in case something is not yaml serializable.
        """

        with FlatDict.open(file) as fp:  # pylint: disable=C0103
            return cls.from_yamls(fp.read(), *args, **kwargs)

    def yamls(self, *args, **kwargs) -> str:
        r"""
        Dump Runner State to yaml string.
        """

        return self.state.yamls(*args, **kwargs)

    @classmethod
    def from_yamls(cls, string: str, *args, **kwargs) -> BaseRunner:
        r"""
        Construct Runner from yaml string.
        """

        return cls(Config.from_yamls(string, *args, **kwargs))

    @property
    def progress(self) -> float:
        r"""
        Training Progress.

        Returns:
            (float):

        Raises:
            RuntimeError: If no terminal is defined.
        """

        return self.steps / self.total_steps

    @property
    def best_fn(self) -> Callable:
        r"""
        Function to determine the best score from a list of scores.

        By default, the `best_fn` returns `min` if `self.state.score_name` is `loss`,
        otherwise, returns `max`.

        Subclass can override this method to accommodate needs, such as `min`.

        Returns:
            (callable):
        """

        return max if self.state.score_name != "loss" else min

    @property
    def best_index(self) -> int:
        r"""
        Find the best index from all scores.

        Returns:
            (int):
        """

        if not self.scores:
            return 0
        values = list(self.scores.values())
        return self.best_fn(range(len(values)), key=values.__getitem__)

    @property
    def latest_result(self) -> NestedDict | None:
        r"""
        Latest result.
        """

        if not self.state.results:
            return None
        latest_index = next(reversed(self.state.results if PY38_PLUS else list(self.state.results)))  # type: ignore
        ret = self.state.results[latest_index].clone()
        ret["index"] = latest_index
        return ret

    @property
    def best_result(self) -> NestedDict | None:
        r"""
        Best result.
        """

        if not self.state.results:
            return None
        best_index = self.best_index
        ret = self.state.results[best_index].clone()
        ret["index"] = best_index
        return ret

    @property
    def scores(self) -> FlatDict | None:
        r"""
        All scores.

        Scores are extracted from results by `score_set` and `runner.state.score_name`,
        following `[r[score_set][self.state.score_name] for r in self.state.results]`.

        Scores are considered as the index of the performance of the model.
        It is useful to determine the best model and the best hyper-parameters.

        `score_set` is defined in `self.state.score_set`.
        If it is not set, `DanLing` will use `val` or `validate` if they appear in the `latest_result`.
        If `DanLing` still could not find, it will fall back to the second key in the `latest_result`
        if it contains more that one element, or the first key.

        Note that certain keys are ignored when falling back, they are defined in {IGNORED_SET_NAMES}.
        """

        if not self.state.results:
            return None
        subsets = [i for i in self.latest_result.keys() if i not in IGNORED_SET_NAMES]  # type: ignore
        score_set = self.state.get("score_set")
        if score_set is None and "val" in subsets:
            score_set = "val"
        if score_set is None and "validate" in subsets:
            score_set = "validate"
        if score_set is None:
            score_set = subsets[1] if len(subsets) > 1 else subsets[0]
        return FlatDict({k: v[score_set][self.state.score_name] for k, v in self.state.results.items()})

    @property
    def latest_score(self) -> float | None:
        r"""
        Latest score.
        """

        if not self.state.results:
            return None
        if not PY38_PLUS:
            return next(reversed(list(self.scores.values())))  # type: ignore
        return next(reversed(self.scores.values()))  # type: ignore

    @property
    def best_score(self) -> float | None:
        r"""
        Best score.
        """

        if not self.state.results:
            return None
        return self.scores[self.best_index]  # type: ignore

    @property
    def is_best(self) -> bool:
        r"""
        If current epoch is the best epoch.
        """

        if not self.state.results:
            return True
        try:
            return abs(self.latest_score - self.best_score) < 1e-7  # type: ignore
        except TypeError:
            return True

    @cached_property
    @ensure_dir
    def dir(self) -> str:
        r"""
        Directory of the run.
        """

        if "dir" in self.state:
            return self.state.dir
        return os.path.join(self.project_root, f"{self.name}-{self.id}")

    @cached_property
    def log_path(self) -> str:
        r"""
        Path of log file.
        """

        if "log_path" in self.state:
            return self.state.log_path
        return os.path.join(self.dir, "run.log")

    @cached_property
    @ensure_dir
    def checkpoint_dir(self) -> str:
        r"""
        Directory of checkpoints.
        """

        if "checkpoint_dir" in self.state:
            return self.state.checkpoint_dir
        return os.path.join(self.dir, self.checkpoint_dir_name)

    def __getattribute__(self, name) -> Any:
        if name in ("__class__", "__dict__"):
            return super().__getattribute__(name)
        if name in dir(self):
            return super().__getattribute__(name)
        if name in self.__dict__:
            return self.__dict__[name]
        if "state" in self and name in self.state:
            return self.state[name]
        return super().__getattribute__(name)

    def __getattr__(self, name) -> Any:
        if "state" not in self:
            raise RuntimeError("Runner is not initialised yet.")
        if name in dir(self.state):
            return getattr(self.state, name)
        return super().__getattribute__(name)

    def __setattr__(self, name, value) -> None:
        if name in self.__dict__:
            if isinstance(self.__dict__[name], Variable):
                self.__dict__[name].set(value)
            else:
                self.__dict__[name] = value
        elif "state" in self and name in self.state:
            if isinstance(self.state[name], Variable):
                self.state[name].set(value)
            else:
                self.state[name] = value
        else:
            object.__setattr__(self, name, value)

    def __contains__(self, name) -> bool:
        return name in dir(self) or ("state" in self.__dict__ and name in dir(self.state))

    def __repr__(self):
        lines = []
        for key, value in self.__dict__.items():
            value_str = repr(value)
            value_str = self._add_indent(value_str)
            lines.append("(" + key + "): " + value_str)

        main_str = self.__class__.__name__ + "("
        if lines:
            main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"
        return main_str

    def _add_indent(self, text):
        lines = text.split("\n")
        # don't do anything for single-line stuff
        if len(lines) == 1:
            return text
        first = lines.pop(0)
        # add 2 spaces to each line but the first
        lines = [(2 * " ") + line for line in lines]
        lines = "\n".join(lines)
        lines = first + "\n" + lines
        return lines

    def init_deepspeed(self, config: dict | None = None) -> dict:  # type: ignore # pylint: disable=R0912,R0915
        r"""
        Preprocess DeepSpeed config.
        """

        if config is None:
            config = self.state.get("deepspeed")
        if config is None:
            return {}
        if isinstance(config, str):
            config = NestedDict.load(config)
        if config.get("steps_per_print", "auto") == "auto":
            config["steps_per_print"] = self.print_interval
        if config.get("train_micro_batch_size_per_gpu", "auto") == "auto":
            config["train_micro_batch_size_per_gpu"] = self.batch_size
        if "amp" in config:
            amp = config["amp"]
            if amp.get("enabled", "auto") == "auto":
                amp["enabled"] = "true"
            if amp.get("opt_level", "auto") == "auto":
                amp["opt_level"] = "O1"
        if "zero_optimization" in config:
            zero = config["zero_optimization"]
            if zero.get("allgather_bucket_size") == "auto":
                zero["allgather_bucket_size"] = 1e6
            if zero.get("reduce_bucket_size") == "auto":
                zero["reduce_bucket_size"] = 1e6
            if zero.get("stage3_max_live_parameters") == "auto":
                zero["stage3_max_live_parameters"] = 1e8
            if zero.get("stage3_max_live_gradients") == "auto":
                zero["stage3_max_live_gradients"] = 1e8
            if zero.get("stage3_max_reuse_distance") == "auto":
                zero["stage3_max_reuse_distance"] = 1e8
            if zero.get("stage3_prefetch_bucket_size") == "auto":
                zero["stage3_prefetch_bucket_size"] = 1e6
            if zero.get("stage3_param_persistence_threshold") == "auto":
                zero["stage3_param_persistence_threshold"] = 1e8
            if "amp" in config:
                if "fp16" not in config:
                    config["fp16"] = {}
                if config["fp16"].get("enabled", "auto"):
                    config["fp16"]["enabled"] = config["amp"]["enabled"]
                warn(
                    f"AMP is not compatible with ZeRO. Automatically set 'fp16' to {config['amp']['enabled']}",
                    stacklevel=2,
                )
                del config["amp"]
        if "optimizer" in config:
            if "params" not in config["optimizer"]:
                config["optimizer"]["params"] = {}
            optimizer = config["optimizer"]["params"]
            if optimizer.get("lr", "auto") == "auto":
                optimizer["lr"] = self.state.get("optim.lr", 1e-3)
            if optimizer.get("weight_decay", "auto") == "auto":
                optimizer["weight_decay"] = self.state.get("optim.weight_decay", 1e-2)
            if optimizer.get("betas") == "auto":
                optimizer["betas"] = (0.9, 0.999)
            if optimizer.get("eps") == "auto":
                optimizer["eps"] = 1e-8
        if "scheduler" in config:
            if "params" not in config["scheduler"]:
                config["scheduler"]["params"] = {}
            scheduler = config["scheduler"]["params"]
            if scheduler.get("total_num_steps", "auto") == "auto":
                scheduler["total_num_steps"] = self.total_steps
            if scheduler.get("warmup_num_steps", "auto") == "auto":
                scheduler["warmup_num_steps"] = scheduler["total_num_steps"] // 20
            if scheduler.get("warmup_max_lr", "auto") == "auto":
                if self.optimizer:
                    scheduler["warmup_max_lr"] = self.optimizer.param_groups[0]["lr"]
                elif "optimizer" in config:
                    scheduler["warmup_max_lr"] = config["optimizer"]["params"]["lr"]
                else:
                    raise ValueError("warmup_max_lr is not defined and cannot be inferred")
            if scheduler.get("warmup_min_lr", "auto") == "auto":
                scheduler["warmup_min_lr"] = 1e-7
        return config

    @on_main_process
    def init_logging(self) -> None:
        r"""
        Set up logging.
        """

        os.makedirs(os.path.dirname(self.log_path), exist_ok=True)
        # Why is setting up proper logging so !@?#! ugly?
        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "formatters": {
                    "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
                },
                "handlers": {
                    "stdout": {
                        "level": "INFO",
                        "formatter": "standard",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                    },
                    "logfile": {
                        "level": "DEBUG",
                        "formatter": "standard",
                        "class": "logging.FileHandler",
                        "filename": self.log_path,
                        "mode": "a",
                    },
                },
                "loggers": {
                    "": {
                        "handlers": ["stdout", "logfile"],
                        "level": "DEBUG",
                        "propagate": True,
                    },
                },
            }
        )
        logging.captureWarnings(True)
        self.logger = logging.getLogger("runner")
        self.logger.flush = lambda: [h.flush() for h in self.logger.handlers]  # type: ignore

    def init_print(self, process: int = 0) -> None:
        r"""
        Set up `print`.

        Only print on a specific `process` or when `force = True`.

        Args:
            process: The process to `print` on.

        Notes
        -----
        If `self.state.log = True`, the default `print` function will be override by `logging.info`.
        """

        logger = logging.getLogger("print")
        logger.flush = lambda: [h.flush for h in logger.handlers]  # type: ignore
        import builtins as __builtin__  # pylint: disable=C0415

        builtin_print = __builtin__.print

        @catch
        def print(*args, force=False, end="\n", file=None, flush=False, **kwargs):  # pylint: disable=W0622
            if self.rank == process or force:
                if self.state.log:
                    logger.info(*args, **kwargs)
                else:
                    builtin_print(*args, end=end, file=file, flush=flush, **kwargs)

        __builtin__.print = print

    @on_main_process
    def init_tensorboard(self, *args, **kwargs) -> None:
        r"""
        Set up Tensoraoard SummaryWriter.
        """
        raise NotImplementedError

    def set_seed(self, seed: int | None = None, bias: int | None = None) -> None:
        r"""
        Set up random seed.

        Args:
            seed: Random seed to set.
                Defaults to `self.state.seed` (`config.seed`).

            bias: Make the seed different for each processes.

                This avoids same data augmentation are applied on every processes.

                Defaults to `self.rank`.

                Set to `False` to disable this feature.
        """

        seed = seed or self.state.seed
        bias = bias or self.rank
        if bias:
            seed += bias
        if np_random is not None:
            np_random.seed(seed)
        random.seed(seed)

    def set_deterministic(self) -> None:
        r"""
        Set up deterministic.
        """

        raise NotImplementedError

    def scale_lr(
        self,
        lr: float,  # pylint: disable=C0103
        lr_scale_factor: float | None = None,
        batch_size_base: int | None = None,
    ) -> float:
        r"""
        Scale learning rate according to [linear scaling rule](https://arxiv.org/abs/1706.02677).
        """

        # pylint: disable=W0201

        if lr_scale_factor is None:
            if batch_size_base is None:
                batch_size_base = getattr(self, "batch_size_base", None)
                if batch_size_base is None:
                    raise ValueError("batch_size_base must be specified to auto scale lr")
            lr_scale_factor = self.batch_size_equivalent / batch_size_base
        elif batch_size_base is not None:
            warn(
                "batch_size_base will be ignored if lr_scale_factor is specified", category=RuntimeWarning, stacklevel=2
            )
        lr = lr * lr_scale_factor  # pylint: disable=C0103, E1101
        self.lr_scale_factor = lr_scale_factor
        return lr

    def step(self, loss, batch_size: int | None = None, zero_grad: bool = True) -> None:
        r"""
        Backward loss and step optimizer & scheduler.

        This method increment `self.state.steps`.

        This method also increment `self.state.iters` when `batch_size` is specified.

        Args:
            zero_grad: Whether to zero the gradients.
        """

        raise NotImplementedError

    def state_dict(self, cls: Callable = dict) -> Mapping:
        r"""
        Return dict of all attributes for checkpoint.
        """

        return cls(self.state)

    @catch
    @on_main_process
    def save_checkpoint(self) -> None:
        r"""
        Save checkpoint to `self.checkpoint_dir`.

        The checkpoint will be saved to `self.checkpoint_dir/latest.pth`.

        If `self.state.save_interval` is positive and `self.state.epochs + 1` is a multiple of `save_interval`,
        the checkpoint will also be copied to `self.checkpoint_dir/epoch-{self.state.epochs}.pth`.

        If `self.is_best` is `True`, the checkpoint will also be copied to `self.checkpoint_dir/best.pth`.
        """

        latest_path = os.path.join(self.checkpoint_dir, "latest.pth")
        self.save(self.state_dict(), latest_path)
        if (
            hasattr(self, "save_interval")
            and self.save_interval > 0
            and (self.state.epochs + 1) % self.save_interval == 0
        ):
            save_path = os.path.join(self.checkpoint_dir, f"epoch-{self.state.epochs}.pth")
            shutil.copy(latest_path, save_path)
        if self.is_best:
            best_path = os.path.join(self.checkpoint_dir, "best.pth")
            shutil.copy(latest_path, best_path)

    def load_checkpoint(  # pylint: disable=W1113
        self,
        checkpoint: Mapping | bytes | str | os.PathLike | None = None,
        auto_resume: bool | None = None,
        override_state: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        Load info from checkpoint.

        Args:
            checkpoint: Checkpoint (or its path) to load.
                Defaults to `self.state.checkpoint`.
            auto_resume: Automatically resume from latest checkpoint if exists.
                Defaults to `False`.
                If is `True` and `checkpoint` is None, will set it to `self.checkpoint_dir/latest.pth`.
            override_state: If True, override runner state with checkpoint state.
                Defaults to `False`.
            *args: Additional arguments to pass to `self.load`.
            **kwargs: Additional keyword arguments to pass to `self.load`.

        Raises:
            FileNotFoundError: If `checkpoint` does not exists.

        See Also:
            [`from_checkpoint`][danling.BaseRunner.from_checkpoint]: Build runner from checkpoint.
            [`load_pretrained`][danling.BaseRunner.load_pretrained]: Load parameters from pretrained checkpoint.
        """

        checkpoint = checkpoint if checkpoint is not None else self.state.get("checkpoint")
        auto_resume = auto_resume if auto_resume is not None else self.state.get("auto_resume", False)

        # TODO: Support loading checkpoints in other format
        if checkpoint is not None:
            if auto_resume:
                warn(
                    "latest checkpoint is preempted by value specified in checkpoint",
                    RuntimeWarning,
                    stacklevel=2,
                )
            if isinstance(checkpoint, (bytes, str, os.PathLike)):
                if not os.path.exists(checkpoint):
                    raise FileNotFoundError(f"checkpoint is set to {checkpoint!r} but does not exist.")
                self.state.checkpoint = checkpoint
                ckpt = self.load(checkpoint, *args, **kwargs)
            elif isinstance(checkpoint, Mapping):
                ckpt = checkpoint
            else:
                raise ValueError(f"pretrained is set to {checkpoint!r} but is not a valid checkpoint.")
        elif auto_resume:
            checkpoint = os.path.join(self.checkpoint_dir, "latest.pth")
            if os.path.exists(checkpoint):
                self.state.checkpoint = checkpoint
                ckpt = self.load(checkpoint, *args, **kwargs)
            else:
                warn("latest checkpoint does not exits", category=RuntimeWarning, stacklevel=2)
                return
        else:
            raise ValueError("checkpoint is not specified and auto_resume is not set to True")

        # TODO: Wrap state_dict in a dataclass
        self.state.merge(ckpt["runner"], overwrite=override_state)
        if self.model is not None and "model" in ckpt:
            model = self.unwrap_model(self.model)
            model.load_state_dict(ckpt["model"])
        if self.optimizer is not None and "optimizer" in ckpt:
            self.optimizer.load_state_dict(ckpt["optimizer"])
        if self.scheduler is not None and "scheduler" in ckpt:
            self.scheduler.load_state_dict(ckpt["scheduler"])
        self.state.iter_begin = self.state.iters
        self.state.step_begin = self.state.steps
        self.state.epoch_begin = self.state.epochs

    @classmethod
    def from_checkpoint(cls, checkpoint: Mapping | bytes | str | os.PathLike, *args, **kwargs) -> BaseRunner:
        r"""
        Build BaseRunner from checkpoint.

        Args:
            checkpoint: Checkpoint (or its path) to load.
            *args: Additional arguments to pass to `cls.load`.
            **kwargs: Additional keyword arguments to pass to `cls.load`.

        Returns:
            (BaseRunner):
        """

        if isinstance(checkpoint, (bytes, str, os.PathLike)):
            ckpt = cls.load(checkpoint, *args, **kwargs)
        elif isinstance(checkpoint, Mapping):
            ckpt = checkpoint
        else:
            raise ValueError(f"checkpoint is set to {checkpoint} but is not a valid checkpoint.")
        runner = cls(**ckpt["runner"])
        runner.load_checkpoint(ckpt, override_state=False)
        return runner

    def load_pretrained(  # pylint: disable=W1113
        self, checkpoint: Mapping | bytes | str | os.PathLike | None = None, *args, **kwargs
    ) -> None:
        """
        Load parameters from pretrained checkpoint.

        This method only loads the model weights.

        Args:
            checkpoint: Pretrained checkpoint (or its path) to load.
                Defaults to `self.state.pretrained`.
            *args: Additional arguments to pass to `self.load`.
            **kwargs: Additional keyword arguments to pass to `self.load`.

        Raises:
            FileNotFoundError: If `checkpoint` does not exists.

        See Also:
            [`load_checkpoint`][danling.BaseRunner.load_checkpoint]: Load info from checkpoint.
        """

        # TODO: Support loading checkpoints in other format
        checkpoint = checkpoint if checkpoint is not None else self.state.get("pretrained")
        if checkpoint is None:
            raise ValueError("pretrained is not specified")
        if isinstance(checkpoint, (bytes, str, os.PathLike)):
            if not os.path.exists(checkpoint):
                raise FileNotFoundError(f"pretrained is set to {checkpoint!r} but does not exist.")
            ckpt = self.load(checkpoint, *args, **kwargs)
        elif isinstance(checkpoint, Mapping):
            ckpt = checkpoint
        else:
            raise ValueError(f"pretrained is set to {checkpoint!r} but is not a valid checkpoint.")
        ckpt = ckpt.get("model", ckpt)
        ckpt = ckpt.get("state_dict", ckpt)
        model = self.unwrap_model(self.model)
        model.load_state_dict(ckpt)

    def append_result(self, result: NestedDict, index: int | None = None) -> None:
        r"""
        Append result to `self.state.results`.

        Warnings:
            `self.state.results` is heavily relied upon for computing metrics.

            Failed to use this method may lead to unexpected behavior.
        """

        if index is None:
            index = self.state.epochs
            global __APPEND_RESULT_COUNTER__
            __APPEND_RESULT_COUNTER__ += 1
            if index == 0 and __APPEND_RESULT_COUNTER__ > 1:
                warn(
                    """
                    Automatically set index to `self.state.epochs`.
                    Please ensure `self.state.epochs` updates before calling `append_result`
                    """,
                    category=RuntimeWarning,
                    stacklevel=2,
                )
        if index in self.state.results:
            self.state.results[index].merge(result)
        else:
            self.state.results[index] = result

    def print_result(self) -> None:
        r"""
        Print latest and best result.
        """

        print(f"results: {self.state.results}")
        print(f"latest result: {self.latest_result}")
        print(f"best result: {self.best_result}")

    def step_log(self, split: str, iteration: int, length: int | None = None):
        if length is None:
            length = len(self.dataloaders[split]) - 1
        result = self.meters.val
        if self.metrics is not None:
            result.merge(self.metrics.val)
        print(self.format_step_result(result, split, iteration, length))
        if self.mode == "train":
            self.write_result(result, split, iteration)
        return result

    def format_step_result(self, result: NestedDict, split: str, steps: int, length: int) -> str:
        result = NestedDict(result).clone()
        repr_str = ""
        if split is not None:
            if self.mode == "train":
                repr_str = f"training on {split} "
            elif self.mode == "eval":
                repr_str = f"evaluating on {split} "
            else:
                repr_str = f"running in {self.mode} mode on {split} "
        repr_str += f"[{steps}/{length}]\t"
        return repr_str + self.format_result(result)

    def format_epoch_result(self, result: NestedDict, epochs: int | None = None, epoch_end: int | None = None) -> str:
        result = NestedDict(result).clone()
        epochs = epochs or self.state.epochs
        epoch_end = epoch_end or self.state.epoch_end
        repr_str = f"epoch [{epochs}/{epoch_end - 1}]\n" if epochs is not None and epoch_end else ""
        repr_str += "\n".join([f"{k}:\t{self.format_result(v)}" for k, v in result.items()])
        return repr_str

    def format_result(self, result):
        return "\t".join([f"{k}: {v}" for k, v in result.items()])

    def write_result(self, result: NestedDict, split: str, steps: int):
        for name, score in result.all_items():
            name = name.replace(".", "/")
            if name == "loss" and isinstance(score, AverageMeter):
                score = score.avg
            if isinstance(score, Sequence):
                for i, s in enumerate(score):
                    self.write_score(f"{name}/{i}", s, split, steps)
            elif isinstance(score, Mapping):
                for k, s in score.items():
                    self.write_score(f"{name}/{k}", s, split, steps)
            else:
                self.write_score(name, score, split, steps)

    def write_score(self, name: str, score: float, split: str, steps: int):
        if self.writer:
            self.writer.add_scalar(f"{split}/{name}", score, steps)  # type: ignore

    @catch
    @on_main_process
    def save_result(self) -> None:
        r"""
        Save result to `self.dir`.

        This method will save latest and best result to
        `self.dir/latest.json` and `self.dir/best.json` respectively.
        """

        results_path = os.path.join(self.dir, "results.json")
        self.save(
            {
                "id": self.state.id,
                "name": self.state.name,
                "results": self.state.results,
            },
            results_path,
            indent=4,
        )
        ret = {"id": self.state.id, "name": self.state.name}
        result = self.latest_result  # type: ignore
        if isinstance(result, FlatDict):
            result = result.dict()  # type: ignore
        # This is slower but ensure id is the first key
        if result is not None:
            ret.update(result)  # type: ignore
        latest_path = os.path.join(self.dir, "latest.json")
        self.save(ret, latest_path, indent=4)
        if self.is_best:
            best_path = os.path.join(self.dir, "best.json")
            shutil.copy(latest_path, best_path)



from chanfig import Config as Config_
from chanfig import NestedDict

import danling as dl


class Runner(dl.BaseRunner):
    def init_distributed(self) -> None:
        pass


class Config(Config_):
    __test__ = False

    def __init__(self):
        self.network.name = "resnet18"
        self.dataset.download = True
        self.dataset.root = "data"
        self.dataloader.batch_size = 8
        self.epoch_end = 2
        self.optim.name = "adamw"
        self.optim.lr = 1e-3
        self.optim.weight_decay = 1e-4
        self.log = False
        self.tensorboard = False
        self.gradient_clip = False
        self.print_interval = 10
        self.train_iterations_per_epoch = 64
        self.val_iterations_per_epoch = 16
        self.index_set = "val"
        self.index = "loss"


class Test:
    config = Config()
    runner = Runner(config)

    def test_results(self):
        runner = self.runner
        state = runner.state
        state.results = NestedDict(
            {
                0: {
                    "val": {
                        "loss": 1.0,
                        "acc": 0.0,
                    },
                },
                1: {
                    "val": {
                        "loss": 0.5,
                        "acc": 0.5,
                    },
                },
                2: {
                    "val": {
                        "loss": 0.2,
                        "acc": 0.8,
                    },
                },
                3: {
                    "val": {
                        "loss": 0.6,
                        "acc": 0.4,
                    },
                },
            }
        )
        assert runner.best_result.dict() == {
            "index": 2,
            "val": {
                "loss": 0.2,
                "acc": 0.8,
            },
        }
        assert runner.latest_result.dict() == {
            "index": 3,
            "val": {
                "loss": 0.6,
                "acc": 0.4,
            },
        }
        assert runner.best_score == 0.2
        assert runner.latest_score == 0.6



from .mlp import MLP, Dense
from .transformer import (
    FullyConnectedNetwork,
    MultiHeadAttention,
    SimpleAttention,
    TransformerDecoder,
    TransformerDecoderLayer,
    TransformerEncoder,
    TransformerEncoderLayer,
    UnitedPositionEmbedding,
)

__all__ = [
    "TransformerEncoder",
    "TransformerEncoderLayer",
    "TransformerDecoder",
    "TransformerDecoderLayer",
    "MultiHeadAttention",
    "SimpleAttention",
    "FullyConnectedNetwork",
    "UnitedPositionEmbedding",
    "MLP",
    "Dense",
]



from __future__ import annotations

from enum import auto
from functools import wraps
from typing import Any

try:
    from enum import StrEnum  # type: ignore # pylint: disable = C0412
except ImportError:
    from strenum import LowercaseStrEnum as StrEnum  # type: ignore


class RunnerMeta(type):
    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        instance = super().__call__(*args, **kwargs)
        instance.__post_init__()
        return instance


class RunnerMode(StrEnum):
    r"""
    `RunnerMode` is an enumeration of running modes.

    Attributes:
        train: Training mode.
        eval: Evaluation mode.
        inf: Inference mode.
    """

    train = auto()
    eval = auto()
    inf = auto()


def on_main_process(func):
    """
    Decorator to run func only on main process.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Any | None:
        if self.is_main_process or not self.distributed:
            return func(self, *args, **kwargs)
        return None

    return wrapper


def on_local_main_process(func):
    """
    Decorator to run func only on local main process.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Any | None:
        if self.is_local_main_process or not self.distributed:
            return func(self, *args, **kwargs)
        return None

    return wrapper



# pylint: disable=E1101,W0622

from __future__ import annotations

from collections.abc import Mapping
from math import nan
from typing import Any, Callable, Iterable

import torch
from chanfig import FlatDict
from torch import Tensor
from torch import distributed as dist
from torcheval.metrics import Metric

from danling.tensors import NestedTensor


def world_size() -> int:
    r"""Return the number of processes in the current process group."""
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size()
    return 1


class flist(list):  # pylint: disable=R0903
    def __format__(self, *args, **kwargs):
        return " ".join([x.__format__(*args, **kwargs) for x in self])


class Metrics(Metric):
    r"""
    Metric class wraps around multiple metrics that share the same states.

    Typically, there are many metrics that we want to compute for a single task.
    For example, we usually needs to compute `accuracy`, `auroc`, `auprc` for a classification task.
    Computing them one by one is inefficient, especially when evaluating in a distributed environment.

    To solve this problem, Metrics maintains a shared state for multiple metric functions.

    Attributes:
        metrics: A dictionary of metrics to be computed.
        input: The input tensor of latest batch.
        target: The target tensor of latest batch.
        inputs: All input tensors.
        targets: All target tensors.

    Args:
        *args: A single mapping of metrics.
        **metrics: Metrics.
    """

    metrics: FlatDict[str, Callable]
    _input: Tensor
    _target: Tensor
    _inputs: list[Tensor]
    _targets: list[Tensor]
    _input_buffer: list[Tensor]
    _target_buffer: list[Tensor]
    score_name: str
    best_fn: Callable
    merge_dict: bool = True

    def __init__(self, *args, merge_dict: bool | None = None, **metrics: FlatDict[str, Callable]):
        super().__init__()
        self._add_state("_input", torch.empty(0))
        self._add_state("_target", torch.empty(0))
        self._add_state("_inputs", [])
        self._add_state("_targets", [])
        self._add_state("_input_buffer", [])
        self._add_state("_target_buffer", [])
        self.metrics = FlatDict(*args, **metrics)
        if merge_dict is not None:
            self.merge_dict = merge_dict

    @torch.inference_mode()
    def update(self, input: Any, target: Any) -> None:
        if isinstance(input, NestedTensor):
            self._input = input
            self._input_buffer.extend(input.to(self.device).storage)
        else:
            if not isinstance(input, torch.Tensor):
                input = torch.tensor(input)
            self._input = input
            self._input_buffer.append(input.to(self.device))
        if isinstance(target, NestedTensor):
            self._target = target
            self._target_buffer.extend(target.to(self.device).storage)
        else:
            if not isinstance(target, torch.Tensor):
                target = torch.tensor(target)
            self._target = target
            self._target_buffer.append(target.to(self.device))

    def compute(self) -> FlatDict[str, float]:
        return self.comp

    def value(self) -> FlatDict[str, float]:
        return self.val

    def average(self) -> FlatDict[str, float]:
        return self.avg

    @property
    def comp(self) -> FlatDict[str, float]:
        return self._compute(self._input, self._target)

    @property
    def val(self) -> FlatDict[str, float]:
        return self._compute(self.input, self.target)

    @property
    def avg(self) -> FlatDict[str, float]:
        return self._compute(self.inputs, self.targets)

    @torch.inference_mode()
    def _compute(self, input: Tensor, target: Tensor) -> flist | float:
        if input.numel() == 0 == target.numel():
            return FlatDict({name: nan for name in self.metrics.keys()})
        ret = FlatDict()
        for name, metric in self.metrics.items():
            score = metric(input, target)
            if isinstance(score, Tensor):
                ret[name] = score.item() if score.numel() == 1 else flist(score.tolist())
            elif isinstance(score, Mapping):
                if self.merge_dict:
                    ret.merge(score)
                else:
                    for n, s in score:
                        ret[f"{name}.{n}"] = s
            else:
                ret[name] = score
        return ret

    @torch.inference_mode()
    def merge_state(self, metrics: Iterable):
        raise NotImplementedError()

    @property
    @torch.inference_mode()
    def input(self):
        if world_size() == 1:
            return self._input
        if isinstance(self._input, Tensor):
            synced_tensor = [torch.zeros_like(self._input) for _ in range(dist.get_world_size())]
            dist.all_gather(synced_tensor, self._input)
            return torch.cat(synced_tensor, 0)
        if isinstance(self._input, NestedTensor):
            synced_tensors = [None for _ in range(dist.get_world_size())]
            dist.all_gather_object(synced_tensors, self._input.storage)
            return NestedTensor([i for j in synced_tensors for i in j])
        raise ValueError(f"Expected input to be a Tensor or a NestedTensor, but got {type(self._input)}")

    @property
    @torch.inference_mode()
    def target(self):
        if world_size() == 1:
            return self._target
        if isinstance(self._target, Tensor):
            synced_tensor = [torch.zeros_like(self._target) for _ in range(dist.get_world_size())]
            dist.all_gather(synced_tensor, self._target)
            return torch.cat(synced_tensor, 0)
        if isinstance(self._target, NestedTensor):
            synced_tensors = [None for _ in range(dist.get_world_size())]
            dist.all_gather_object(synced_tensors, self._target.storage)
            return NestedTensor([i for j in synced_tensors for i in j])

    @property
    @torch.inference_mode()
    def inputs(self):
        if not self._inputs and not self._input_buffer:
            return torch.empty(0)
        if self._input_buffer:
            if world_size() > 1:
                synced_tensors = [None for _ in range(dist.get_world_size())]
                dist.all_gather_object(synced_tensors, self._input_buffer)
                self._inputs.extend([i for j in synced_tensors for i in j])
            else:
                self._inputs.extend(self._input_buffer)
            self._input_buffer = []
        if isinstance(self._input, NestedTensor):
            return NestedTensor(self._inputs)
        return torch.cat(self._inputs, 0)

    @property
    @torch.inference_mode()
    def targets(self):
        if not self._targets and not self._target_buffer:
            return torch.empty(0)
        if self._target_buffer:
            if world_size() > 1:
                synced_tensors = [None for _ in range(dist.get_world_size())]
                dist.all_gather_object(synced_tensors, self._target_buffer)
                self._targets.extend([i for j in synced_tensors for i in j])
            else:
                self._targets.extend(self._target_buffer)
            self._target_buffer = []
        if isinstance(self._target, NestedTensor):
            return NestedTensor(self._targets)
        return torch.cat(self._targets, 0)

    def __repr__(self):
        keys = tuple(i for i in self.metrics.keys())
        return f"{self.__class__.__name__}{keys}"

    def __format__(self, format_spec):
        val, avg = self.compute(), self.average()
        return "\n".join(
            [f"{key}: {val[key].__format__(format_spec)} ({avg[key].__format__(format_spec)})" for key in self.metrics]
        )


class ScoreMetrics(Metrics):
    r"""
    `ScoreMetrics` is a subclass of Metrics that supports scoring.

    Score is a single value that best represents the performance of the model.
    It is the core metrics that we use to compare different models.
    For example, in classification, we usually use auroc as the score.

    `ScoreMetrics` requires two additional arguments: `score_name` and `best_fn`.
    `score_name` is the name of the metric that we use to compute the score.
    `best_fn` is a function that takes a list of values and returns the best value.
    `best_fn` is only not used by `ScoreMetrics`, it is meant to be accessed by other classes.

    Attributes:
        score_name: The name of the metric that we use to compute the score.
        best_fn: A function that takes a list of values and returns the best value.

    Args:
        *args: A single mapping of metrics.
        score_name: The name of the metric that we use to compute the score. Defaults to the first metric.
        best_fn: A function that takes a list of values and returns the best value. Defaults to `max`.
        **metrics: Metrics.
    """

    score_name: str
    best_fn: Callable

    def __init__(
        self, *args, score_name: str | None = None, best_fn: Callable | None = max, **metrics: FlatDict[str, Callable]
    ):
        super().__init__(*args, **metrics)
        self.score_name = score_name or next(iter(self.metrics.keys()))
        self.metric = self.metrics[self.score_name]
        self.best_fn = best_fn or max

    def score(self, scope: str) -> float | flist:
        if scope == "batch":
            return self.batch_score()
        if scope == "average":
            return self.average_score()
        raise ValueError(f"Unknown scope: {scope}")

    def batch_score(self) -> float | flist:
        return self.calculate(self.metric, self.input, self.target)

    def average_score(self) -> float | flist:
        return self.calculate(self.metric, self.inputs, self.targets)



from danling import metrics, modules, optim, registry, runner, tensors, typing, utils

from .metrics import (
    AverageMeter,
    AverageMeters,
    Metrics,
    binary_metrics,
    multiclass_metrics,
    multilabel_metrics,
    regression_metrics,
)
from .registry import GlobalRegistry, Registry
from .runner import BaseRunner, TorchRunner
from .tensors import NestedTensor, PNTensor
from .utils import catch, debug, ensure_dir, flexible_decorator, is_json_serializable, load, method_cache, save

__all__ = [
    "metrics",
    "modules",
    "optim",
    "registry",
    "runner",
    "tensors",
    "utils",
    "typing",
    "BaseRunner",
    "TorchRunner",
    "Registry",
    "GlobalRegistry",
    "Metrics",
    "AverageMeter",
    "AverageMeters",
    "regression_metrics",
    "binary_metrics",
    "multiclass_metrics",
    "multilabel_metrics",
    "NestedTensor",
    "PNTensor",
    "save",
    "load",
    "catch",
    "debug",
    "flexible_decorator",
    "method_cache",
    "ensure_dir",
    "is_json_serializable",
]



import sys
from contextlib import contextmanager
from typing import Optional

from danling.typing import Exceptions

try:
    import ipdb as pdb
except ImportError:
    import pdb  # type: ignore


@contextmanager
def debug(
    enable: bool = True,
    error: Exceptions = Exception,
    exclude: Optional[Exceptions] = None,
):
    """
    Contextmanager to enter debug mode on `error` except for `exclude`.

    `debug` is intended to be used to catch the error and enter debug mode.
    Since it is mainly for development purposed, we include an `enable` args so that it can be deactivated.

    Args:
        enable: Whether to enable the contextmanager.
            Defaults to `True`.
        error: The error to catch.
            Defaults to `Exception`.
        exclude: The error to exclude.
            Defaults to `None`.
    """

    if not enable:
        yield
        return
    try:
        yield
    except error as exc:  # pylint: disable=W0703
        if exclude is not None and isinstance(exc, exclude):
            raise exc
        _, m, tb = sys.exc_info()  # pylint: disable=C0103
        print(m.__repr__(), file=sys.stderr)  # pylint: disable=C2801
        pdb.post_mortem(tb)
    finally:
        pass



from __future__ import annotations

from chanfig import DefaultDict, NestedDict


class AverageMeter:
    r"""
    Computes and stores the average and current value.

    Attributes:
        val: Current value.
        avg: Average value.
        sum: Sum of values.
        count: Number of values.

    Examples:
    ```python
    >>> meter = AverageMeter()
    >>> meter.update(0.7)
    >>> meter.val
    0.7
    >>> meter.avg
    0.7
    >>> meter.update(0.9)
    >>> meter.val
    0.9
    >>> meter.avg
    0.8
    >>> meter.sum
    1.6
    >>> meter.count
    2
    >>> meter.reset()
    >>> meter.val
    0
    >>> meter.avg
    0

    ```
    """

    val: float = 0
    avg: float = 0
    sum: float = 0
    count: int = 0

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        r"""
        Resets the meter.

        Examples:
            >>> meter = AverageMeter()
            >>> meter.update(0.7)
            >>> meter.val
            0.7
            >>> meter.avg
            0.7
            >>> meter.reset()
            >>> meter.val
            0
            >>> meter.avg
            0
        """

        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n: int = 1) -> None:
        r"""
        Updates the average and current value in the meter.

        Args:
            val: Value to be added to the average.
            n: Number of values to be added.

        Examples:
            >>> meter = AverageMeter()
            >>> meter.update(0.7)
            >>> meter.val
            0.7
            >>> meter.avg
            0.7
            >>> meter.update(0.9)
            >>> meter.val
            0.9
            >>> meter.avg
            0.8
            >>> meter.sum
            1.6
            >>> meter.count
            2
        """

        # pylint: disable=C0103

        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __format__(self, format_spec) -> str:
        return f"{self.val.__format__(format_spec)} ({self.avg.__format__(format_spec)})"


class AverageMeters(DefaultDict):
    r"""
    A `DefaultDict` for `AverageMeter`.

    Examples:
    ```python
    >>> meters = AverageMeters()
    >>> meters.loss.reset()
    >>> meters.loss.update(0.7)
    >>> meters.loss.val
    0.7
    >>> meters.loss.avg
    0.7
    >>> meters.update(0.9)
    >>> meters.loss.val
    0.9
    >>> meters.loss.avg
    0.8
    >>> meters.loss.sum
    1.6
    >>> meters.loss.count
    2
    >>> meters.reset()
    >>> meters.loss.val
    0
    >>> meters.loss.avg
    0

    ```
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs.setdefault("default_factory", AverageMeter)
        super().__init__(*args, **kwargs)

    @property
    def val(self) -> NestedDict[str, float]:
        return NestedDict({key: meter.val for key, meter in self.items()})

    @property
    def avg(self) -> NestedDict[str, float]:
        return NestedDict({key: meter.avg for key, meter in self.items()})

    @property
    def sum(self) -> NestedDict[str, float]:
        return NestedDict({key: meter.sum for key, meter in self.items()})

    @property
    def count(self) -> NestedDict[str, int]:
        return NestedDict({key: meter.count for key, meter in self.items()})

    def reset(self) -> None:
        r"""
        Resets all meters.

        Examples:
            >>> meters = AverageMeters()
            >>> meters.loss.update(0.7)
            >>> meters.loss.val
            0.7
            >>> meters.loss.avg
            0.7
            >>> meters.reset()
            >>> meters.loss.val
            0
            >>> meters.loss.avg
            0
        """

        for meter in self.values():
            meter.reset()

    def update(self, val, n: int = 1) -> None:  # type: ignore # pylint: disable=W0237
        r"""
        Updates the average and current value in all meters.

        Args:
            val: Value to be added to the average.
            n: Number of values to be added.

        Note:
            This function is **NOT** recommended to use, as it alters all meters in the bank.

        Examples:
            >>> meters = AverageMeters()
            >>> meters.loss.update(0.7)
            >>> meters.loss.val
            0.7
            >>> meters.loss.avg
            0.7
            >>> meters.update(0.9)
            >>> meters.loss.val
            0.9
            >>> meters.loss.avg
            0.8
            >>> meters.loss.sum
            1.6
            >>> meters.loss.count
            2
        """

        for meter in self.values():
            meter.update(val, n)

    def __format__(self, format_spec) -> str:
        return "\n".join(f"{key}: {meter.__format__(format_spec)}" for key, meter in self.items())
